/* $Id: cpp.c,v 1.5 2000/11/27 20:09:44 malekith Exp $ */

#include "util.h"
#include "phash.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdarg.h>

#define CHUNK 1024

char *cpp_file;
int cpp_line;

/* 
 * this describes either {regular file, then name is set, line has
 * usual meaning and fd points to result of open(),} or {virtual
 * "file" - a string from expansion, name == 0, line is undefinied
 * and fd == -1}
 */
struct cpp_file {
	int fd;
	byte *buf;
	int len, pos;
	int line;
	char *name;
};

enum {
	s_line_beg,
	s_word_beg,
	s_norm,
	s_string,
};

struct cpp_state {
	struct pstack *file_stack;
	struct dbuf *db_path;
	struct phash *symbols;
	struct cpp_file *current_file;
	int vfile;
	int state;
	int s_delim;
	int skip;		/* level of ifs in skip mode (ifnot) */
	int lev;		/* level of #ifs */
	char *lookahead;
	char buf[256];
};

enum {
	sym_simple,
};

struct cpp_symbol {
	struct cpp_symbol *next;
	char *name;
	
	int kind;
	int last_used;
	char *val;
};


#define CTX struct cpp_state

static CTX *cpp_ctx;

static void vmsg(CTX *ctx, const char *fmt, va_list ap)
{
	(void)ctx;

	fprintf(stderr, "%s:%d: ", cpp_file, cpp_line);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
}

static void dead_msg(CTX *ctx, const char *fmt, ... )
{
	va_list ap;

	va_start(ap, fmt);
	vmsg(ctx, fmt, ap);
	va_end(ap);

	exit(1);
}

static void msg(CTX *ctx, const char *fmt, ... )
{
	va_list ap;

	va_start(ap, fmt);
	vmsg(ctx, fmt, ap);
	va_end(ap);
}

static char *concat(const char *l, const char *r)
{
	char *x;

	x = malloc(strlen(l) + strlen(r) + 1);
	strcpy(x, l);
	strcat(x, r);
	return x;
}

static inline int read_more(struct cpp_file *f)
{
	if (f->len == -1)
		return -1;

	if (f->pos >= f->len) {
		if (f->fd == -1)
			f->len = -1;
		else
			f->len = read(f->fd, f->buf, CHUNK);
		if (f->len <= 0) {
			f->len = -1;
			return -1;
		}
		f->pos = 0;
	}

	if (f->buf[f->pos] == '\n')
		f->line++;
		
	return f->buf[f->pos++];
}

/*
 * check c for special meaning
 * return what had to be returned by higher level function
 * or -2 to restart.
 */
static inline int check_spec(struct cpp_file *f, int c)
{
	int c2;

	if (c != '/' && c != '\\')
		return c;
		
	c2 = read_more(f);

	if (c == '\\' && c2 == '\n')
		return -2;		/* made get() read once more char */
					/* so that \\ \n is eaten */

	if (c != '/') 
		goto exit;

	switch (c2) {
	case '*':
		while ((c = read_more(f)) != -1) 
			while (c == '*') 
				if ((c = read_more(f)) == '/')
					return ' ';
		return -1;
	case '/':
		while ((c = read_more(f)) != -1)
			if (c == '\n')
				break;
		return c;
	}

exit:
	/* unget and do nothing with c */
	f->pos--;
	return c;
}

static inline int file_get(struct cpp_file *f)
{
	int r;

	do {
		r = read_more(f);
		r = check_spec(f, r);
	} while (r == -2);
	
	return r;
}

static inline void set_global_vars(CTX *ctx)
{
	if (ctx->current_file && ctx->current_file->fd != -1) {
		cpp_file = ctx->current_file->name;
		cpp_line = ctx->current_file->line;
	}
}

static void close_file(struct cpp_file *f)
{
	if (f->fd != -1)
		close(f->fd);
	free(f->buf);
	free(f);
	/* DON'T free f->name, it's needed later on ! */
}

static void open_vfile(CTX *ctx, const char *caption, int last)
{
	struct cpp_file *f;

	f = NEW(struct cpp_file);
	memset(f, 0, sizeof(*f));
	f->fd = -1;
	f->len = strlen(caption);
	if (last != -1)
		f->len++;
	f->buf = (byte*)malloc(f->len + 1);
	strcpy((char*)f->buf, caption);
	if (last != -1) {
		*(strchr((char*)f->buf, 0) + 1) = 0;
		*(strchr((char*)f->buf, 0)) = last;
	}

	/*
	 * vfile is incressed on each expanded identifier in user
	 * supplaied data (i.e. NOT in strings resulting from
	 * expanding other id's).
	 * this is to stop inifinite recusion, symbols have info
	 * where (in which vfile) they were last used.
	 */
	if (ctx->current_file->fd != -1)
		ctx->vfile++;
	push(ctx->file_stack, ctx->current_file);
	ctx->current_file = f;
}

static void check_double(CTX *ctx, char *x)
{
	struct cpp_file **f, **e;

	f = (struct cpp_file**)ctx->file_stack->beg;
	e = (struct cpp_file**)ctx->file_stack->sp;

	for (; f < e; f++)
		if ((*f)->name && strcmp((*f)->name, x) == 0) 
			dead_msg(ctx, "file `%s' included recursivly", x);
}

static void open_file(CTX *ctx, const char *fn, int main, int look_here)
{
	struct cpp_file *f;
	int fd;
	char **p, *x;
	int n;

	if (main) {
		x = strdup(fn);
		fd = open(x, O_RDONLY);
		if (fd == -1) {
			perror(fn);
			exit(1);
		}
	} else {
		if (*fn == '/' || look_here) {
			x = strdup(fn);
			fd = open(x, O_RDONLY);
			if (fd != -1)
				goto found;
			free(x);
		}
		
		n = db_off(ctx->db_path);
		n /= sizeof(char*);
		p = (char**)db_data(ctx->db_path);

		while (n--) {
			x = concat(*p++, fn);
			fd = open(x, O_RDONLY);
			if (fd != -1)
				break;
			free(x);
		}
		
		if (fd == -1) 
			dead_msg(ctx, "kpp: cannot find #include file '%s'", 
				fn);

	found:
		check_double(ctx, x);
	}
	
	f = NEW(struct cpp_file);
	memset(f, 0, sizeof(*f));
	f->name = x;
	f->fd = fd;
	f->buf = (byte*)malloc(CHUNK);
	f->line = 1;

	if (ctx->current_file)
		push(ctx->file_stack, ctx->current_file);
	ctx->current_file = f;
	set_global_vars(ctx);
}

static int get(CTX *ctx)
{
	int r;
	
	if (ctx->current_file == 0)
		return -1;

	do {
		r = file_get(ctx->current_file);
		
		if (r == -1) {
			close_file(ctx->current_file);
			ctx->current_file = 0;
			if (pstack_empty(ctx->file_stack))
				return -1;
			ctx->current_file = (struct cpp_file*)
				pop(ctx->file_stack);
			set_global_vars(ctx);
		}
	} while (r == -1);

	return r;
}

static int is_idchar(int c)
{
	return isalnum(c) || c == '_';
}

static void set_lookahead(CTX *ctx, int c)
{
	ctx->buf[0] = c;
	ctx->buf[1] = 0;
	ctx->lookahead = ctx->buf;
}

static int id(CTX *ctx, int c)
{
	int i = 0;
	int last = -1;
	struct cpp_symbol *s;
	
	for (ctx->buf[i++] = c; i < 250; i++) {
		c = get(ctx);
		if (!is_idchar(c)) {
			last = c;
			break;
		}
		ctx->buf[i] = c;
	}
	ctx->buf[i] = 0;

	if ((s = ph_find(ctx->symbols, ctx->buf)) && 
	    (s->last_used != ctx->vfile || ctx->current_file->fd != -1)) {
		open_vfile(ctx, s->val, last);
		s->last_used = ctx->vfile;
		set_lookahead(ctx, 0);
		return -2;
	}

	/* 
	 * so it wasn't #define, restart with state = s_norm.
	 */
	ctx->state = s_norm;
	if (last != -1)
		ctx->buf[i++] = last;
	ctx->buf[i] = 0;
	ctx->lookahead = ctx->buf;
	return -2;
}

static int get_look(CTX *ctx)
{
	if (*ctx->lookahead)
		return *ctx->lookahead++;

	return get(ctx);
}


struct directive {
	const char *name;
	int (*func)(CTX*, char*);
	int flags;
};

enum {
	f_args		= 0x0001,
	f_can_args	= 0x0002,
	f_no_args	= 0x0004,
	f_if_kind	= 0x0008,
	f_endif_kind	= 0x0010,
	f_else_kind	= 0x0020,
};

static void skip_mode(CTX *ctx);

static int check_defs(CTX *ctx, char *args, int all)
{
	char *e;

	while (*args) {
		while (*args == ' ' || *args == '\t')
			args++;
		for (e = args; *e; e++)
			if (*e == ' ' || *e == '\t')
				break;
		if (*e)
			*e++ = 0;
		if (ph_find(ctx->symbols, args)) {
			if (!all)
				return 1;
		} else {
			if (all)
				return 0;
		}
		args = e;
	}

	return all;
}

static int include(CTX *ctx, char *args)
{
	char *e;
	int look_here = 0;

	while (isspace(*args))
		args++;

	switch (*args) {
	case '"':
	case '<':
		e = strchr(args + 1, *args);
		if (e == 0) 
			dead_msg(ctx, "unmatched `%c' in #include", *args);
		look_here = *args == '"';
		*e++ = 0;
		++args;
		break;
	case 0:
		dead_msg(ctx, "#include requires parameter");
	default:
		dead_msg(ctx, "#include expanded files not supported at "
			"the moment");
	}

	open_file(ctx, args, 0, look_here);
	
	return -2;
}

static int endif(CTX *ctx, char *args)
{
	if (ctx->lev == 0)
		dead_msg(ctx, "unmatched #endif");

	ctx->lev--;

	(void)args;
	
	return -2;
}

static int ifndef(CTX *ctx, char *args)
{
	int r;

	r = check_defs(ctx, args, 0);
	if (!r)
		ctx->lev++;
	else
		skip_mode(ctx);

	return -2;
}

static int ifdef_all(CTX *ctx, char *args)
{
	int r;

	r = check_defs(ctx, args, 1);
	if (r)
		ctx->lev++;
	else
		skip_mode(ctx);

	return -2;
}

static int ifdef_any(CTX *ctx, char *args)
{
	int r;

	r = check_defs(ctx, args, 0);
	if (r)
		ctx->lev++;
	else
		skip_mode(ctx);

	return -2;
}

static int xelse(CTX *ctx, char *args)
{
	(void)args;

	if (ctx->lev == 0) 
		dead_msg(ctx, "unmatched #else");
	ctx->lev--;
	skip_mode(ctx);
	return -2;
}

static int undef(CTX *ctx, char *args)
{
	struct cpp_symbol *s;
	char *e;
	
	while (*args == ' ' || *args == '\t')
		args++;
	for (e = args; *e; e++)
		if (*e == ' ' || *e == '\t')
			break;
	*e = 0;
	
	s = ph_unlink(ctx->symbols, args);

	if (s) {
		free(s->name);
		free(s->val);
		free(s);
	}
	
	return -2;
}

static int warn(CTX *ctx, char *args)
{
	msg(ctx, "#warning %s", args);
	return -2;
}

static int error(CTX *ctx, char *args)
{
	dead_msg(ctx, "#error %s", args);
	return -2;
}

static int define(CTX *ctx, char *args)
{
	struct cpp_symbol *s;
	char *n;
	
	while (*args == ' ' || *args == '\t')
		args++;
	n = args;
	while (is_idchar(*args))
		args++;
	if (*args)
		*args++ = 0;
	s = NEW(struct cpp_symbol);
	memset(s, 0, sizeof(*s));
	s->kind = sym_simple;
	s->name = strdup(n);
	s->val = strdup(args);
	ph_add(ctx->symbols, s);
	return -2;
}

struct directive directives[] = {
	{ "include", include, f_args },
	{ "define", define, f_args },
	{ "undef", undef, f_args },
	{ "ifdef", ifdef_all, f_args|f_if_kind },
	{ "ifndef", ifndef, f_args|f_if_kind },
	{ "endif", endif, f_no_args|f_endif_kind },
	{ "ifdef_any", ifdef_any, f_args|f_if_kind },
	{ "ifdef_all", ifdef_all, f_args|f_if_kind },
	{ "else", xelse, f_no_args|f_else_kind },
	{ "warn", warn, f_can_args },
	{ "warning", warn, f_can_args },
	{ "error", error, f_can_args },
	{ "err", error, f_can_args },
	{ 0, 0, 0 }
};

static int match_directive(CTX *ctx, char **aargs)
{
	int i = 0, c;
	char *args;
	struct dbuf *db;
	
	set_global_vars(ctx);

	do {
		c = get(ctx); 
	} while (c == ' ' || c == '\t');
	
	if (c == '\n' || c == -1) { 
		msg(ctx, "empty '#' found");
		return -1;
	}
	
	ctx->buf[i++] = c;
	while (i < 250) {
		c = get(ctx);
		if (!is_idchar(c)) 
			break;
		ctx->buf[i++] = c;
	}
	ctx->buf[i] = 0;

	for (i = 0; directives[i].name; i++)
		if (strcmp(ctx->buf, directives[i].name) == 0)
			break;

	if (c == '\n' || c == -1)
		args = 0;
	else {
		db = db_new();
		db_addc(db, c);
		while ((c = get(ctx)) != '\n')
			db_addc(db, c);
		db_addc(db, 0);
		args = db_finish(db, 0);
	}

	if (aargs)
		*aargs = args;
	else if (args)
		free(args);
		
	if (directives[i].name == 0) {
		msg(ctx, "unknown #directive `%s'", ctx->buf);
		set_lookahead(ctx, 0);
		return -1;
	}

	set_lookahead(ctx, 0);
	return i;
}

static void skip_mode(CTX *ctx)
{
	int c, start, i;
	char *start_f;
	
	ctx->skip++;
	start_f = cpp_file;
	start = cpp_line;
	
	while (ctx->skip) {
		do c = get(ctx); while (c == ' ' || c == '\t');

		if (c != '#') {
			while (c != -1 && c != '\n')
				c = get(ctx);
			if (c == -1) {
				cpp_file = start_f;
				cpp_line = start;
				dead_msg(ctx, "cannot find matching "
					"#endif");
			}
		} else {
			i = match_directive(ctx, 0);
			if (directives[i].flags & f_if_kind)
				ctx->skip++;
			else if (directives[i].flags & f_endif_kind)
				ctx->skip--;
			else if (directives[i].flags & f_else_kind)
				if (ctx->skip == 1) {
					ctx->skip--;
					ctx->lev++;
				}
		}
	}
}

static int directive(CTX *ctx)
{
	int i = 0, r = -2;
	char *args;

	i = match_directive(ctx, &args);
	if (i < 0)
		return -2;

	if (args == 0 && (directives[i].flags & f_args)) 
		dead_msg(ctx, "#%s needs param(s)",
			directives[i].name);
	else
		r = directives[i].func(ctx, args);

	if (args)
		free(args);

	set_lookahead(ctx, 0);
	
	return r;
}

/*
 * process regular input, return -2 if set lookahead, met string or
 * simply need to restart. -1 on EOF. char otherwise.
 */
static int get_norm(CTX *ctx)
{
	int c;

	c = get_look(ctx);
	if (c == -1)
		return -1;
		
	if (ctx->state == s_line_beg && c == '#')
		return directive(ctx);
		
	if ((ctx->state == s_line_beg || ctx->state == s_word_beg) &&
	    is_idchar(c))
	    	return id(ctx, c);

	switch (c) {
	case '\n':
		ctx->state = s_line_beg;
		break;
	case ' ':
	case '\t':
		if (ctx->state == s_norm)
			ctx->state = s_word_beg;
		break;
	case '"':
	case '\'':
		ctx->state = s_string;
		ctx->s_delim = c;
		break;
	default:
		if (is_idchar(c))
			ctx->state = s_norm;
		else
			ctx->state = s_word_beg;
		break;
	}
	
	return c;
}

static int x_get(CTX *ctx)
{
	int c;

restart:		
	switch (ctx->state) {
	case s_line_beg:
	case s_word_beg:
	case s_norm:
		c = get_norm(ctx);
		if (c == -2)
			goto restart;
		return c;
	case s_string:
		c = get_look(ctx);
		if (c == '\\') {
			c = get(ctx);
			if (c != -1)
				set_lookahead(ctx, c);
			c = '\\';
		} else if (c == ctx->s_delim)
			ctx->state = s_norm;
		return c;
	default:
		halt();
	}
}

int cpp_get()
{
	int c;

	c = x_get(cpp_ctx);
	set_global_vars(cpp_ctx);
	return c;
}

void cpp_init(const char *fn)
{
	CTX *ctx;

	ctx = NEW(CTX);
	memset(ctx, 0, sizeof(*ctx));
	
	ctx->file_stack = pstack_new();
	ctx->symbols = ph_new();
	ctx->vfile = 1;
	ctx->buf[0] = 0;
	ctx->lookahead = ctx->buf;
	
	ctx->db_path = db_new();
	db_addp(ctx->db_path, "./");
	
	open_file(ctx, fn, 1, 0);

	cpp_ctx = ctx;
}

#if 0
int main(int argc, char **argv)
{
	int c;
	
	if (argc < 2) {
		fprintf(stderr, "USAGE: %s filename\n", argv[0]);
		exit(1);
	}
	
	cpp_init(*++argv);
	while ((c = cpp_get()) != -1)
		putchar(c);
	putchar('\n');
	exit(0);
}
#endif
